--[[-----------------------------------------------------------------------------
-- Ardent Defender is a standalone addon for Paladins that provides several features
handy for tanks. 

** Track and Announce Ardent Defender saving your life
** Alerts you when another person causes a parryhaste on your target
** Also alerts if you cause a parryhaste on a boss someone else is tanking.
** Uses libDataBroker to give you damage and ardent save stats for current session
** Announces bubblewall UP and DOWN status. (No more making macros!)
** Announce if your taunt fails and why (IMMUNE, Resist) and if your avenger shield misses
** Tells you some statistic information about your defenses (or lack thereof),
** Allows quick marking of targets by holding down ALT and mousing over targets.

--** Thanks to: 
	Ace3, for the framework.
	hypehuman from WoWHead comments for his formula for calculating AD Heals
	All my friends and guildmates on US-Thrall for helping me test this addon.
	Tankadin2 - Ardent Defender plugin for the idea and some of the original code (now deprecated and replaced)
	

---  All Rights Reserved
---- I alone reserve the rights to distribution of this addon. It may not be copied or distributed on any other websites without my express written permission.
---- WowInterface.com and Curse.com are the only websites permitted to distribute my addons.
--------------------------------------------------------------------------------]]

-- Starts us off. Thank you Ace3!
local addonName = "Ardent Defender"
ArdentDefender = LibStub("AceAddon-3.0"):NewAddon("ArdentDefender", "AceConsole-3.0", "AceEvent-3.0")
local addon = ArdentDefender

addon.version = GetAddOnMetadata("ArdentDefender", "Version")
addon.debug = false
addon.enabled = true


local AD_FLASHTEXTURE = "Interface\\Addons\\ArdentDefender\\Media\\flash.tga" -- Flashes screen blue
local AD_SOUND = "Sound\\Spells\\SimonGame_Visual_BadPress.wav" -- Ahh.. brings back memories of BC, doesn't it?

local function IsProt()
	local currentSpec = GetSpecialization()
	local currentSpecName = (currentSpec and select(2, GetSpecializationInfo(currentSpec)) or "None")

	if (currentSpecName == "Protection") then
		return true
	else
		return false
	end
end

local function DmgMitigated()
    local baseVal = 30
    local SotR = GetMastery()+baseVal
    return SotR
end

-- Gives Total Avoidance
local function totalAvoidance()
	return 5 + (GetDodgeChance() + GetParryChance())
end

-- Gives ideal Parry vs Current Dodge
local function GetIdealParry()
    local d=GetDodgeChance()
    local p=(235.5*d/65.631440)-((235.5/65.631440)*5.01-3.76)
    return p,d
end

-- Gives current ratio between Parry vs Dodge, should aim for 2,5 if possible
local function currRatioAvoid()
    local d=GetDodgeChance()
    local r=((235.5/65.631440)-((235.5/65.631440)*5.01-3.76)/d)
    return r 
end

function addon:handleInterrupt(target,spellid)
	
	if (addon.db.profile.announceInterrupt == false) then
		return
	end
	local text = format("I interrupted %s's %s ", target,GetSpellLink(spellid))
	addon:Output(text,addon.db.profile.tauntOutput,"none")
	
end
local ANCIENTGUARDIAN = 86150
local RaidTags = {  [0] = nil, [1] = nil, [2] = nil, [3] = nil, [4] = nil, [5] = nil, [6] = nil, [7] = nil, [8] = nil}

local ArdentSpells= {
	Other = {
		[6940] = "HANDSAC",
		[1022] = "HOP",
		[1038] = "SALV",
		[48788] = "LOH",
		[633] = "LOH",
		[2800] = "LOH",
		[10310] = "LOH",
		[27154] = "LOH",
		[19753] = "DI",
		[1044] = "FREEDOM",
		[70940] = "DIVGUARD",
		--[86657] = "ANCIENTGUARDIAN",
		
	},
	Self = {
		--[64205] = "DIVSAC",defunc
		
		--[66233] = "AVERT", -- Ardent Defender proc (de)buff defunc
		[498] = "BUBBLEWALL", 
		[66235] = "ADHEAL",
		[31935] = "AVENGERMISS", -- all ranks of avenger shield
		[62124] = "TAUNTMISS", -- Hand of reckoning
		[31789] = "TAUNTMISS", -- Righteous Defense
		[25780] = "RFHACK", -- Little hack to keep an eye on RF
		[31850] = "ARDENTDEFENDER",
		[31821] = "DEVAURA",
	},
	Taunts = {
	[355] = "WARRIORTAUNT", -- Taunt
	[694] = "WARRIORTAUNT", -- Mocking Blow
	[1161] = "WARRIORTAUNT", -- Challenging shout
	[49576] = "DKTAUNT", -- Death grip
	[56222] = "DKTAUNT", -- Dark Command
	[62124] = "PALLYTAUNT", -- Hand of reckoning
	--[31789] = "PALLYTAUNT", -- Righteous Defense -- Deprecated 5.04
	[6795] = "DRUIDTAUNT", -- Growl
	[5209] = "DRUIDTAUNT", -- Challenging Roar
	[20736] = "HUNTERTAUNT", -- Distracting shot
	},
}

-- I don't use most of these colors, but I might someday!
local WHT = "|cffffffff";
local BLK = "|cff000000";
local YEL = "|cffffff20";
local RED = "|cffff2020";
local GRN = "|cff20ff20";
local BLU = "|cff2020ff";
local HIL = "|cfffbd284";

--------------------------------------------------------------------------------
-- Sets up our basic options and defaults using AceConfig-3.0
--------------------------------------------------------------------------------

local function AddonOptions() 
return  {
	name = "Ardent Defender",
	handler = ArdentDefender,
	type = "group",
	get = function(info) return ArdentDefender.db.profile[info[#info]] end,
	set = function(info, value) ArdentDefender.db.profile[info[#info]] = value end,

	args = {
		general = {
			type = "group",
			name = "General Settings",
			cmdInline = true,
			order = 1,
			args = {
				info = {
					type = "description",
					order = 1,
					name = "Provides several useful tools to help weary Tankadin's in their travels.\nIf you are a Paladin and love taking hits.. this addon is for you!\n\nEnjoy.\n-Blinddate, US-Thrall.                Version ".. ArdentDefender.version,
				},
				generalheader = {
					order = 2,
					type = "header",
					name = "General Options",
				},
				nulloption = {
					order = 4,
					type = "description",
					name = " ",
					cmdHidden = true
				},
				
				combatAnnounce = {
					type = "toggle", order = 6.5, 
					name = "End of Combat stats",
					desc = "Toggles displaying combat stats when exiting combat. Will only display if you took damage.",
				},
				parryTracker = {
					type = "select", order = 6,
					name = "Parry Tracker",
					desc = "Choose setting for parry tracking. Whisper, Local only, or disabled.",
					values = { ["WHISPER"] = "WHISPER", ["LOCAL"] = "LOCAL", ["DISABLED"] = "DISABLED" },
				},
				avengerTracker = {
					type = "select", order = 5,
					name = "Avenger Misses",
					desc = "Choose channel for Avenger Misses. Local, disabled, Party, or Raid",
					values = { ["LOCAL"] = "LOCAL", ["PARTY"] = "PARTY", ["DISABLED"] = "DISABLED", ["RAID"] = "RAID"},
				},
				announceInterrupt = {
					type = "toggle", order = 9.5,
					name = "Interrupt Announcer",
					desc = "If enabled, announces when you interrupt an enemy spellcast. Use Taunt channel for output.",
				},
				onlyTanking = {
					type = "toggle", order =9.6,
					name = "Tanking Only",
					desc = "If enabled, announcements are only made when you are currently spec'd for tanking.",
				},
				tauntWatcher = {
					type = "toggle", order = 9,
					name = "Taunt failure tracking",
					desc = "Announces taunt failures due to miss, resist or immunities.\n",
				},
				specialFX = {
					type = "toggle", order = 7,
					name = "Special Effects",
					desc = "Toggles the use of effects such as screen flashing and sounds when AD procs.",
				},
				annoyRF = {
					type = "toggle",order = 8,
					name = "Righteous Fury",
					desc = "Flashes screen constantly if you are in tank spec do not have the RF buff.",
				},
				otherTaunt = {
					type = "toggle", order = 9,
					name = "Announce Others' Taunts",
					width = "full",
					desc = "If enabled, announces to channel when someone taunts your target. If disabled, it announces it localy only.",
				},
				nulloption = {
					order = 9,
					type = "description",
					name = " ",
					cmdHidden = true
				},
				quietMode = {
					type ="toggle",
					name = "Quiet mode",
					desc = "Quiet mode keeps the addon enabled, but all output is sent to local chat window.",
				},
				basicMode = {
					type = "toggle",
					name = "Basic mode",
					desc = "If enabled, all output is filtered through the Taunt Channel, and all other output options are disabled. Hand of ... are WHISPERED not announced.",
				},
				quickMarks = {
					type = "toggle",
					name = "Quick Target Marker",
					desc = "If enabled, holding ALT while you mouseover targets will mark them in order. Skull, X, Square, etc.",
					set = function(info,value) 
						addon.db.profile.quickMarks = value
						ArdentDefender:ResetTags()
						end,
					get = function(info,value) return addon.db.profile.quickMarks end,
				},
				tauntinfo = {
					order = 10.5,
					type = "multiselect",
					name= "Taunt Channel",
					desc = "The channel(s) selected here will be used to announce when someone\n Taunts your current tank target, or when your Taunt misses, is resisted, or target is immune to taunt.",
					values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING"},
						set = function(info,key,value) addon.db.profile.tauntOutput[key] = value end,
						get = function(info,value) return addon.db.profile.tauntOutput[value] end,
					},
				},
			},
		messages = {
			name = "Custom Messages",
			type = "group",
			childGroups = "tab",
			order = 2,
			args = {
				bubbleWall= {
					order = 3,
					type = "group",
					name = "Bubblewall settings",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,

					args = {
					
						messageUp = {
							type = "input", width = "full", order = 1,
							name = "Bubblewall Active.",
							desc = "Message to be announced when your bubblewall is active.",
							usage ="Type your message.",
							
						},
					
						messageDown = {
							type ="input", width = "full", order = 2,
							name = "Bubblewall Down",
							desc = "Message to be announced when your bubblewall is down.",
							usage = "Type your message.",
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
						
				},
					ardentDefender= {
					order = 3,
					type = "group",
					name = "Ardent Defender settings",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,

					args = {
					
						messageUp = {
							type = "input", width = "full", order = 1,
							name = "Ardent Defender Active.",
							desc = "Message to be announced when your Ardent Defender is active.",
							usage ="Type your message.",
							
						},
					
						messageDown = {
							type ="input", width = "full", order = 2,
							name = "Adent Defender Down",
							desc = "Message to be announced when your Ardent Defender is down.",
							usage = "Type your message.",
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
						
				},

				devAura = {
					order = 4,
					type = "group",
					name = "Devotion Aura Settings",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,
					
					args = {
					
						messageUp = {
							type ="input", width = "full",cmdHidden = true, order = 4,
							name = "Devotion Aura Active",
							desc = "Message to be announced when your Devotion Aura is used.",
							usage = "Type your message.",
						},
						messageDown = {
							type ="input", width = "full",cmdHidden = true, order = 5,
							name = "Devotion Aura faded",
							desc = "Message to be announced when your Devotion Aura fades.",
							usage = "Type your message.",
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
				},
				ancientGuardian = {
					order = 5,
					type = "group",
					name = "Ancient Guardian Settings",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,
					
					args = {
					
						messageUp = {
							type ="input", width = "full",cmdHidden = true, order = 4,
							name = "Ancient Guardian Active",
							desc = "Message to be announced when you have an Ancient Guardian protecting you.",
							usage = "Type your message.",
						},
						--messageDown = {
						---	type ="input", width = "full",cmdHidden = true, order = 5,
						--	name = "Ancient Guardian down",
						--	desc = "Message to be announced when your Ancient Guardian is gone.",
						--	usage = "Type your message.",
						--},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
				},
				avengerAvert = {
					type = "group",
					name = "Avenger / Avert",
					desc = "Avenger Shield / Avert settings",
					
					args = {
					
						avengerMessage = {
							type ="input", width = "full",order = 1,
							name = "Avenger Shield Missed",
							desc = "Message announced when your Avenger Shield misses one or more targets. ",
							usage = "Type your message.",
							set = function(info,value) addon.db.profile.avengerMessage = value end,
							get = function(info,value) return addon.db.profile.avengerMessage end,
						},
						avertMessage = {
							type = "input", order = 1, width = "full",
							name = "Death avoided message.",
							desc = "The message to be displayed when Ardent Defender procs and your life.\n",
							usage = "Type your message. <Heal> in your message will be replaced with how much Ardent Defender healed you for, if it healed you.",
							set = function(info,value) addon.db.profile.avert.messageUp = value end,
							get = function(info,key) return addon.db.profile.avert.messageUp end,
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING"},
							set = function(info,key,value) addon.db.profile.avert.output[key] = value end,
							get = function(info,value) return addon.db.profile.avert.output[value] end,
						},
					},
				},
			},
		},
		statistics = {
			name = "Statistics",
			type = "group",
			order = 4,
			args = {
				info = {
					type = "description",
					order = 1,
					name = function() return format ("You have Dodged %d Attacks.\r\nYou have also Parried %d times.\r\nAnd your enemies have outright MISSED you %d times.\r\nFurther more, you have avoided %d deaths, blocked %d damage,\r\n and absorbed %d damage.\r\nPlease note these statistics are reset each time you login or reset your UI.",
					AD_Info.session.hitsDodged or 0, AD_Info.session.hitsParried or 0, AD_Info.session.hitsMissed or 0, AD_Info.session.DeathsAverted or 0 , AD_Info.session.DmgBlocked or 0 , AD_Info.session.DmgAbsorbed or 0) end,
					},
			},
		},
		handmessages = {
			name = "Hand of ... messages",
			
			type = "group",
			childGroups = "tab",
			order = 3,
			args = {
				handSac = {
					type = "group",
					name = "Sacrifice",
					desc = "Settings for Hand of Sacrifice",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,
					
					args = {
						messageUp = {
							type ="input", width = "full", order = 1,
							name = "Hand of Sacrifice message active",
							desc = "Message announced when your Hand of Sacrifice is activated.",
							usage = "Type your message. <target> will be replaced with the name of the player you placed your hand on.",
						},
						messageDown = {
							type ="input", width = "full",order = 2,
							name = "Hand of Sacrifice down",
							desc = "Message announced when your Hand of Sacrifice has ended.",
							usage = "Type your message. <you> is replaced by target name if announced to raid.",
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING", ["WHISPER"] = "WHISPER"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
				},
				handSalv = {
					type = "group",
					name = "Salvation",
					desc = "Settings for Hand of Salvation",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,
					args = {
						messageUp = {
							type ="input", width = "full",cmdHidden = true, order = 1,
							name = "Hand of Salvation active",
							desc = "Message WHISPERED to target when you use Hand of Salvation on them.",
							usage = "Type your message. <you> is replaced by target name if announced to raid.",
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING", ["WHISPER"] = "WHISPER"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
				},
				handProt = {
					type = "group",
					name = "Protection",
					desc = "Settings for Hand of Protection",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,
					args = {
						messageUp = {
							type ="input", width = "full",cmdHidden = true, order = 1,
							name = "Hand of Protection active",
							desc = "Message WHISPERED to target when you use Hand of Protection on them.",
							usage = "Type your message. <you> is replaced by target name if announced to raid.",
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING", ["WHISPER"] = "WHISPER"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
				},
				handfree = { 
					type = "group",
					name = "Freedom",
					desc = "Settings for Hand of Freedom",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,
					args = {
						messageUp = {
							type ="input", width = "full",cmdHidden = true, order = 1,
							name = "Hand of Freedom message",
							desc = "Message WHISPERED when you use Hand of Freedom on them.",
							usage = "Type your message. <you> is replaced by target name if announced to raid.",
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING", ["WHISPER"] = "WHISPER"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
				},
				loh = {
					type = "group",
					name = "Lay on Hands",
					desc = "Settings for Lay on Hands",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,
					args = {
						messageUp = {
							type ="input", width = "full",cmdHidden = true, order = 1,
							name = "Lay On Hands",
							desc = "Message WHISPERED to target when you use your Lay on Hands on them.",
							usage = "Type your message. <you> is replaced by target name if announced to raid.",
						},
						output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING", ["WHISPER"] = "WHISPER"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
				},
				di = {
					type = "group",
					name = "Divine Intervention",
					desc = "Settings for Divine Intervention",
					set = function(info,value) addon.db.profile[info[#info-1]][info[#info]] = value end,
					get = function(info,key) return addon.db.profile[info[#info-1]][info[#info]] end,
					args = {
						messageUp = {
							type = "input", width  = "full", order = 1,
							name = "Divine Intervention",
							desc= "Message WHISPERED when you use Divine Intervention on someone.",
							usage ="Type your message",
						}, 
					output = {
							hidden = function() return (addon.db.profile.basicMode == true) end,
							name = "Output channel",
							desc= "Channel used for output on this ability.",
							order = 3,
							type = "multiselect",
							values = { ["SAY"] = "SAY",["YELL"] = "YELL",["LOCAL"] = "LOCAL",["PARTY"] = "PARTY",["RAID"] = "RAID",["RAID_WARNING"] = "RAID_WARNING", ["WHISPER"] = "WHISPER"},
							set = function(info,key,value) addon.db.profile[info[#info-1]][info[#info]][key] = value end,
							get = function(info,value) return addon.db.profile[info[#info-1]][info[#info]][value] end,
						},
					},
				},
			}
		},
		
	}

}
end
--------------------------------------------------------------------------------
-- Standard defaults for config data.
--------------------------------------------------------------------------------
local defaults = {
    profile =  {
		avert = {
			messageUp = "Holy #$% did you see that? I got healed for <heal> thanks to Ardent Defender!!!",
			messageDown = nil,
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = true,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = false,
			},
		},
		ardentDefender = {
			messageUp = "Ardent Defender ACTIVE. I liked the old version better. :/",
			messageDown = "Ardent Defender is downnn. Just like Paladin class representation.",
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = true,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = false,
			},
		},
		
		
		bubbleWall = {
			messageUp = "Bubblewall ACTIVE!!",
			messageDown = "Bubblewall is down.",
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = true,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = false,
			},
		},
		devAura = {
			messageUp = "\124cff71d5ff\124Hspell:31821\124h[Devotion Aura]\124h\124r is ACTIVE!! All magic damage reduced 20%!",
			messageDown = "\124cff71d5ff\124Hspell:31821\124h[Devotion Aura]\124h\124r is down",
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = true,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = false,
			},
		},
		handSac = {
				messageUp = "\124cff71d5ff\124Hspell:6940\124h[Hand of Sacrifice]\124h\124r ACTIVE on <target>, please watch my health!!",
				messageDown = "\124cff71d5ff\124Hspell:6940\124h[Hand of Sacrifice]\124h\124r down.",
				output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = false,
				["PARTY"] = true,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = false,
			},
		},
		handSalv = {
			messageUp = "(Auto) I used my \124cff71d5ff\124Hspell:1038\124h[Hand of Salvation]\124h\124r on <you>.",
			messageDown = nil,
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = false,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = true,
			},
		},
		handProt = {
				messageUp = "(Auto) I used my \124cff71d5ff\124Hspell:10278\124h[Hand of Protection]\124h\124r on <you>!.",
				messageDown = nil,
				output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = false,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = true,
			},
		},
		di = {
			messageUp = "(Auto) I used my \124cff71d5ff\124Hspell:19753\124h[Divine Intervention]\124h\124r on <you>. Right click the buff to dispell it if you need to.",
			messageDown = nil,
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = false,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = true,
			},
		},
		loh = {
			messageUp = "(Auto) I used my \124cff71d5ff\124Hspell:48788\124h[Lay on Hands]\124h\124r on <you>!",
			messageDown = nil,
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] =false,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = true,
			},
		},
		handfree = {
			messageUp = "(Auto) I used my \124cff71d5ff\124Hspell:1044\124h[Hand of Freedom]\124h\124r on <you>!",
			messageDown = nil,
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = false,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = true,
			},
		},
		tauntOutput = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = false,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = false,
		},
		ancientGuardian = {
			messageUp = "I have a Guardian Angel!",
			messageDown = "My angel has left me.. :(",
			output = {
				["SAY"] = false,
				["YELL"] = false,
				["LOCAL"] = true,
				["PARTY"] = false,
				["RAID"] = false,
				["RAID_WARNING"] = false,
				["WHISPER"] = false,
			},
		},
			
		avengerMessage = "Avenger shield missed!! Watch your threat!",
		avengerTracker = "PARTY",
		
		combatAnnounce = false, -- Announce end of combat info in chatbox. Default off.
		tauntWatcher = true, -- Announces when your taunt target is IMMUNE, it resists, or it misses. Also watches Avenger Shield
		specialFX = true, -- Screen shaking and flashing
		annoyRF = false,
		otherTaunt = false,
		quietMode = false,
		basicMode = true,
		announceInterrupt = true,
		onlyTanking = true,
		quickMarks = false,
		
		parryTracker = "DISABLED",
		
		version = 2.0, -- Just incase I need to change crap in the future...
		
				
	},

}


--------------------------------------------------------------------------------
-- Sets up our local variables, session and last combat data.
-- This data does not persist, resets every time you logout or reset the UI.
--------------------------------------------------------------------------------
AD_Info = {
		adProc = 0,
		session = {
			hitsParried = 0,
			hitsMissed = 0,
			hitsDodged = 0,
			DamageTaken = 0,
			DeathsAverted = 0,
			SpellDmgTaken = 0,
			DmgBlocked = 0,
			AmtHealed = 0,
			DmgAbsorbed = 0,
		},
		lastCombat = {
			hitsParried = 0,
			hitsMissed = 0,
			hitsDodged = 0,
			DamageTaken = 0,
			DeathsAverted = 0,
			SpellDmgTaken = 0,
			DmgBlocked = 0,
			AmtHealed = 0,
			DmgAbsorbed = 0,
    	},
}

--------------------------------------------------------------------------------
-- OnInitialize()
-- Initialization function, sets up options table, database, slash commands, LDB Object,etc.
--------------------------------------------------------------------------------
function ArdentDefender:OnInitialize()

	local _, class = UnitClass("player")
	if (class ~= "PALADIN") then
		addon:Debug("Not a paladin.")
		return
	end
	
    -- Initialize the database including defaults.    
	addon.db = LibStub("AceDB-3.0"):New("ArdentDefenderDB", defaults, "Default")

	-- Setup the options table.
    LibStub("AceConfig-3.0"):RegisterOptionsTable("ArdentDefender", AddonOptions())
    addon.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ArdentDefender", "ArdentDefender")

	-- Register our slash commands
    addon:RegisterChatCommand("ad", "ChatCommand")
    addon:RegisterChatCommand("ardentdefender", "ChatCommand")

    	-- LibDataBroker Object.Displays overall session info, ability to open config window, reset data and announce session data.
	addon.ldb = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("Ardent Defender", {
		type = "launcher",
		label = "AD",
		OnClick = function(_, msg)
			if (msg == "LeftButton") then
				if (IsShiftKeyDown()) then
					local partynum = GetNumSubgroupMembers()
					
					if (IsInRaid() == true) then
						ArdentDefender:FormatStats("SESSION","RAID")
					elseif (not IsInRaid()and partynum > 0) then
						ArdentDefender:FormatStats("SESSION","PARTY")
					else
						ArdentDefender:Print(ArdentDefender:FormatStats("SESSION",nil))
					end

				else
					InterfaceOptionsFrame_OpenToCategory("ArdentDefender")
				end
			elseif (msg == "RightButton" and IsAltKeyDown()) then
				ArdentDefender:ClearData()
			end
		end,
			icon = "Interface\\Icons\\Spell_Holy_ArdentDefender",
		OnTooltipShow = function(tooltip)
			if not tooltip or not tooltip.AddLine then return end
				
						
			local mastery = GetCombatRatingBonus(CR_MASTERY)
			local level = UnitLevel("player")
			
			tooltip:AddLine(format("|rActive Mitigation:\n"))			
--			tooltip:AddLine(format("%s|r\n",BLU))
			
			--tooltip:AddLine(format("%sArdent Defender|r",WHT))
			tooltip:AddLine(format("%s~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|r\n",BLU))
			
			
			local adHeal = math.floor((UnitHealthMax("player") * 0.15))
			local adText = RED.."Ardent Defender is not available.|r"
			local SotRText = RED.."Shield of the Righteous is not available.|r"
			
			if (IsProt() and level >=70)  then -- We have both, great!
				adText = format("%sAD will heal: |r%d",WHT,adHeal)
				SotRText = format("%sSotR Mitigation: |r%d%%",WHT,DmgMitigated())
    			elseif (IsProt() and level >=40 and level < 70)  then -- We have SotR but not AD
				SotRText = format("%sSotR Mitigation: |r%d%%",WHT,DmgMitigated())
    			end

			tooltip:AddDoubleLine(adText,SotRText)
		                
			tooltip:AddLine("|r\n\nPassive Mitigation:\n")
--		          tooltip:AddLine(format("%s|r\n",BLU))
			
			local idealParry, currDodge = GetIdealParry() -- Gives ideal Parry vs Current Dodge
			--ArdentDefender:Debug("Ideal parry ratio: "..tostrong(idealParry).. " "..tostrong(currDodge))
			tooltip:AddLine(format("%s\nYour parry should be: |r%.2f%%. %s(vs |r%.2f%% %sDodge)|r",WHT,idealParry,WHT,currDodge,WHT))

                        local currRatio = currRatioAvoid() -- Gives current ratio between Parry vs Dodge, should aim for 2,5 if possible
			tooltip:AddLine(format("%sCurrent Ratio: |r%.2f. %s(2.5 is the 'ideal' ratio)|r",WHT,currRatio,WHT))
			
			-- Block Removed, now it's on separate roll, after Avoidance roll
			
			tooltip:AddLine(format("%sTotal avoidance: |r%.2f%%%s.|r\n\nChance to be hit:\n",WHT,totalAvoidance(),WHT))
			local blockVal = GetBlockChance()
			local parryVal = GetParryChance()
			local dodgeVal = GetDodgeChance()
			local missVal = 3
			local missChances = { [0] = nil, [1] = nil, [2] = nil, [3] = nil }
			
			for i = 0, 3 do
				local block = max(blockVal - 1.5 * i, 0)
				local parry = max(parryVal - 1.5 * i, 0)
				local dodge = max(dodgeVal - 1.5 * i, 0)

        	    		local miss = max(missVal - 1.5 * i, 0) -- Will be 0 for levels +2/3
                
				local hitChance = max(100.0-parry-dodge-miss, 0) -- Non-avoided attacks (overall hit + block)
				local calcBlock = min(hitChance*block/100, hitChance) -- Chance to block over all attacks (displayed block % is chance to block over attacks that are not avoided)
				local calcHit = hitChance-calcBlock -- Chance to be hit over all attacks (100%-avoidance-converted block)
			
				local level = level+i; -- Mob Level        

				if (i == 3) then
					level = level.." / |TInterface\\TargetingFrame\\UI-TargetingFrame-Skull:0|t";
				end
			--	tooltip:AddLine(format("%sby Level %s: |r%.2f%%%s.|r",WHT,level,calcHit,WHT))
				missChances[i] = format("%sby Level %s: |r%.2f%%%s.|r",WHT,tostring(level),calcHit,WHT)
				
			end            
			tooltip:AddDoubleLine(missChances[0],missChances[1])
			tooltip:AddDoubleLine(missChances[2],missChances[3])
							
			tooltip:AddLine(addon:FormatStats("SESSION",nil))
			
			tooltip:AddLine("|cffeda55fClick|r to show options menu.",0,1,0)
			tooltip:AddLine("|cffeda55fAlt-Right-Click|r to clear all data.",0,1,0)
			tooltip:AddLine("|cffeda55fShift-Click to report data to party/raid channel",0,1,0)
	end, })

	addon:Debug("Finished loading.")
end


--------------------------------------------------------------------------------
-- addon:Debug(...)
-- Simple debug function. If Debug is enabled, prints message to player chatbox
--------------------------------------------------------------------------------
function ArdentDefender:Debug(...)
	if (addon.debug) then
		addon:Print("Debug:",...)
	end
end

--------------------------------------------------------------------------------
-- OnEnable()
-- Called when Addon is loaded(?) Register our Events, etc.
--------------------------------------------------------------------------------
function ArdentDefender:OnEnable()

	
	local _, class = UnitClass("player")
	if (class ~= "PALADIN") then
		addon:Debug("Not a paladin.")
		return
	end
	
	addon.enabled = true
	addon:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	addon:RegisterEvent("PLAYER_REGEN_DISABLED") -- Entering combat
 	addon:RegisterEvent("PLAYER_REGEN_ENABLED") -- Exiting combat
	addon:RegisterEvent("PLAYER_ENTERING_WORLD") -- Game loaded
	addon:RegisterEvent("UPDATE_MOUSEOVER_UNIT") 

		
end
--------------------------------------------------------------------------------
-- OnDisable()
-- Called when the Addon is disabled, unregisters our events.
--------------------------------------------------------------------------------
function ArdentDefender:OnDisable()

	local _, class = UnitClass("player")
	if (class ~= "PALADIN") then
		addon:Debug("Not a paladin.")
		return
	end
	addon.enabled = false
	addon:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	addon:UnregisterEvent("PLAYER_REGEN_DISABLED")
	addon:UnregisterEvent("PLAYER_REGEN_ENABLED")
	addon:UnregisterEvent("PLAYER_ENTERING_WORLD")
	addon:UnregisterEvent("UPDATE_MOUSEOVER_UNIT") 
end

--------------------------------------------------------------------------------
-- Our command only serves to open the config menu at the moment. 
--------------------------------------------------------------------------------
function ArdentDefender:ChatCommand(input)
	if not input or input:trim() == "" then
		InterfaceOptionsFrame_OpenToCategory(addon.optionsFrame)
	elseif (input == "standby") then
		if (addon.enabled) then
			addon:Print("Ardent Defender DISABLED. Type /ad standby to resume.")
			addon:Disable()
		elseif (addon.enabled == false) then
			addon:Print("Ardent Defender ENABLED.")
			addon:Enable()
		end
	elseif (input == "quiet") then
		if (addon.db.profile.quietMode == true) then
			addon:Print("Quiet Mode disabled.")
			addon.db.profile.quietMode = false
		else
			addon:Print("Quiet mode enabled.")
			addon.db.profile.quietMode = true
		end
	elseif (input == "reset" or input == "reset tags") then
			ArdentDefender:ResetTags()
			addon:Print("All marked tags reset.")
			ArdentDefender:ResetTags()
			SetRaidTargetIcon("player",0)
		
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(ArdentDefender, "ad", "ArdentDefender", input)
	end
end

	
--------------------------------------------------------------------------------
-- addon:ClearData()
-- Clears all saved data from session and combat Useful.. I guess?
--------------------------------------------------------------------------------
function ArdentDefender:ClearData()

	AD_Info.session.DamageTaken = 0
	AD_Info.session.DeathsAverted = 0
	AD_Info.session.AmtHealed = 0
	AD_Info.session.DmgBlocked = 0
	AD_Info.session.DmgAbsorbed = 0
	AD_Info.session.hitsParried = 0
	AD_Info.session.hitsDodged = 0
	AD_Info.session.hitsMissed = 0
	AD_Info.adProc = 0


	AD_Info.lastCombat.DamageTaken = 0
	AD_Info.lastCombat.DeathsAverted = 0
	AD_Info.lastCombat.AmtHealed = 0
	AD_Info.lastCombat.DmgBlocked = 0
	AD_Info.lastCombat.DmgAbsorbed = 0
	AD_Info.lastCombat.hitsParried = 0
	AD_Info.lastCombat.hitsDodged = 0
	AD_Info.lastCombat.hitsMissed = 0

	ArdentDefender:Print("All saved data cleared.")
 end


 local function isMember(flags) 
 
	if (bit.band(flags, COMBATLOG_OBJECT_AFFILIATION_PARTY) > 0) then
		return true
	elseif (bit.band(flags,COMBATLOG_OBJECT_AFFILIATION_RAID) > 0) then
		return true
	elseif (bit.band(flags,COMBATLOG_OBJECT_AFFILIATION_MINE) > 0) then
		return true
	else
		return false
	end
end
	
--------------------------------------------------------------------------------
-- addon:COMBAT_LOG_EVENT_UNFILTERED(...)
-- Has a Godaweful amount of parameters, depending on which event is triggered. I hate this.
-- Combat log handler. Checks for Parrys, Bubblewall, and Ardent Defender Heals, Divine Sacrifice, etc etc..
--------------------------------------------------------------------------------
function ArdentDefender:COMBAT_LOG_EVENT_UNFILTERED(...)
	local eventType, _,sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destRaidFlags , _, prefix1, prefix2, prefix3 = select(3,...)
	local suffixPos = 16
	
	local PlayerGUID = UnitGUID("player")
	local apply = (eventType == "SPELL_AURA_APPLIED" and true or false)
	
	
	
	if ( eventType == "SPELL_AURA_APPLIED" or eventType == "SPELL_AURA_REMOVED") then
		if (sourceGUID == PlayerGUID and destGUID ~=PlayerGUID and ArdentSpells.Other[prefix1]) then
			addon:Announce(ArdentSpells.Other[prefix1],apply,destName)
		elseif (sourceGUID == PlayerGUID and destGUID == PlayerGUID and ArdentSpells.Self[prefix1]) then
			addon:Announce(ArdentSpells.Self[prefix1],apply,destName)
--		elseif (destGUID == PlayerGUID and prefix1 == 86657) then
--			addon:Announce("ANCIENTGUARDIAN", apply, destName)
		-- Okay nasty hack for Divine Guardian
		--elseif (sourceGUID ~= PlayerGUID and ArdentSpells.Other[prefix1] == "DEVAURA" and apply == false and isMember(sourceFlags)) then
			--addon:Announce("DEVAURA",false,destName)
		end
	elseif (eventType == "UNIT_DIED" or eventType == "PARTY_KILL" and addon.db.profile.quickMarks) then
		for i = 1, 8 do
			if (RaidTags[i] == destGUID) then
				RaidTags[i] = nil
			end
			
			
		end
			
	elseif (eventType == "SWING_MISSED") then
		
		
		if(prefix1 == "ABSORB" and destGUID == PlayerGUID) then -- We ABSORBED an attack fully, let's record it.
			AD_Info.lastCombat.DmgAbsorbed = AD_Info.lastCombat.DmgAbsorbed + prefix3
			AD_Info.session.DmgAbsorbed = AD_Info.session.DmgAbsorbed + prefix3
			--addon:Debug("DmgAbsorbed: ",prefix2)
		elseif (prefix1 == "BLOCK" and destGUID == PlayerGUID) then -- We BLOCKED an attack fully, let's record the amount.
			AD_Info.lastCombat.DmgBlocked = AD_Info.lastCombat.DmgBlocked + prefix3
			AD_Info.session.DmgBlocked = AD_Info.session.DmgBlocked + prefix3
			--addon:Debug("Dmgblocked: ", prefix2)
		elseif (prefix1 == "PARRY" ) then
			if (destGUID == PlayerGUID ) then 
				AD_Info.session.hitsParried = 1 + AD_Info.session.hitsParried 
				AD_Info.lastCombat.hitsParried = 1 + AD_Info.lastCombat.hitsParried
			end
			if (UnitGUID("targettarget") == PlayerGUID and sourceGUID ~= PlayerGUID and destGUID == UnitGUID("target")) then -- When someone else causes the boss to parry while you are tanking
				--addon:Debug("ParryOther: ",sourceName,sourceFlags,"/",bit.band(sourceFlags,COMBATLOG_OBJECT_TYPE_PLAYER))
				addon:Announce("OTHERPARRY", (bit.band(sourceFlags, COMBATLOG_OBJECT_TYPE_PLAYER) ~=0),sourceName) -- the bit_band returns true or false (hopefuly) of whether it was a PLAYER or NPC who caused the parry.
			elseif (UnitGUID("targettarget") ~= PlayerGUID and sourceGUID == PlayerGUID)  then -- This is when you are DPSing and cause a parry.
				addon:Announce("PARRY",apply,destName)
			end
		elseif (prefix1 == "MISS" and destGUID == PlayerGUID) then
			AD_Info.session.hitsMissed = 1 +  AD_Info.session.hitsMissed
			AD_Info.lastCombat.hitsMissed = 1 +  AD_Info.lastCombat.hitsMissed
		elseif (prefix1 == "DODGE" and destGUID == PlayerGUID) then
			AD_Info.session.hitsDodged = 1 +  AD_Info.session.hitsDodged
			AD_Info.lastCombat.hitsDodged = 1 +  AD_Info.lastCombat.hitsDodged
		end
	-- New: Check for Taunt and Avenger shield misses
	elseif ( eventType == "SPELL_MISSED" and sourceGUID == PlayerGUID and ArdentSpells.Self[prefix1]) then 
		local immunity = select(suffixPos,...)
		addon:Announce(ArdentSpells.Self[prefix1],(immunity == "IMMUNE" and true or false),destName)
	elseif (eventType == "SPELL_CAST_SUCCESS" and sourceGUID ~=PlayerGUID and destGUID == UnitGUID("target") and UnitGUID("targettarget") == PlayerGUID and ArdentSpells.Taunts[prefix1]) then
		addon:Announce("OTHERTAUNT",prefix2,sourceName)
	-- Check for Ardent Defender heal
	elseif (eventType == "SPELL_CAST_SUCCESS" and sourceGUID == PlayerGUID and prefix1 == ANCIENTGUARDIAN	and IsProt()) then
		addon:Announce("ANCIENTGUARDIAN",true,sourceName)
		
	elseif (eventType == "SPELL_HEAL" and sourceGUID == PlayerGUID and destGUID == PlayerGUID and ArdentSpells.Self[prefix1] and ArdentSpells.Self[prefix1] == "ADHEAL") then -- We detected an Ardent Defender heal, lets record it.
		local amount = select(suffixPos,...)
		-- Seriously.. Why does LUA not have a simple increment function? Why can't this be just a ++ like in C?
		AD_Info.lastCombat.AmtHealed = AD_Info.lastCombat.AmtHealed + amount
		AD_Info.session.AmtHealed = AD_Info.session.AmtHealed + amount
		AD_Info.lastCombat.DeathsAverted = 1 + AD_Info.lastCombat.DeathsAverted
		AD_Info.session.DeathsAverted = 1 + AD_Info.session.DeathsAverted
		addon:Announce("AVERT",true,amount)
	
	elseif ((eventType == "SPELL_DAMAGE" or eventType == "SWING_DAMAGE" or eventType == "RANGE_DAMAGE") and destGUID == PlayerGUID) then -- Player receives spell dmg
		local dmg, overkill, school, resisted, blocked, absorbed = 0
		if (eventType == "SWING_DAMAGE" or eventType == "RANGE_DAMAGE") then
			dmg, overkill, school, resisted, blocked, absorbed = select(suffixPos-3,...)
		else
			dmg, overkill, school, resisted, blocked, absorbed = select(suffixPos,...)
		end

		AD_Info.lastCombat.DamageTaken = AD_Info.lastCombat.DamageTaken + dmg
		AD_Info.session.DamageTaken = AD_Info.session.DamageTaken + dmg
		--addon:Debug("DMGEvent: ", eventType, "Dmg in: ", dmg)
		if (blocked and blocked > 0) then
		    AD_Info.lastCombat.DmgBlocked = AD_Info.lastCombat.DmgBlocked + blocked
		    AD_Info.session.DmgBlocked = AD_Info.session.DmgBlocked + blocked
		    --addon:Debug("DMGEvent: Dmg: ",dmg, " Blocked: ", blocked)
		end
		if (absorbed and absorbed > 0) then
		    AD_Info.lastCombat.DmgAbsorbed = AD_Info.lastCombat.DmgAbsorbed + absorbed
		    AD_Info.session.DmgAbsorbed = AD_Info.session.DmgAbsorbed + absorbed
		    --addon:Debug("DMGEvent: Dmg: ", dmg, " Absorbed: ", absorbed)
		end
	elseif (eventType == "SPELL_HEAL" and sourceGUID == PlayerGUID and destGUID ~= PlayerGUID and ArdentSpells.Other[prefix1] == "LOH") then
		addon:Announce("LOH",true,destName)
	elseif (eventType == "SPELL_INTERRUPT" and sourceGUID == PlayerGUID) then
		local spellID = select(suffixPos,...)
		addon:handleInterrupt(destName,spellID)
	end
	
end

--------------------------------------------------------------------------------
-- addon:PLAYER_REGEN_DISABLED()
-- Triggered when entering combat. Clears combat values, checks talents just incase.
--------------------------------------------------------------------------------
function ArdentDefender:PLAYER_REGEN_DISABLED()

	AD_Info.lastCombat.DamageTaken = 0
 	AD_Info.lastCombat.DeathsAverted = 0
 	AD_Info.lastCombat.AmtHealed = 0
 	AD_Info.lastCombat.DmgBlocked = 0
 	AD_Info.lastCombat.DmgAbsorbed = 0
	AD_Info.lastCombat.hitsParried = 0
	AD_Info.lastCombat.hitsMissed= 0
	AD_Info.lastCombat.hitsDodged = 0
	AD_Info.adProc = 0
end

--------------------------------------------------------------------------------
-- addon:PLAYER_REGEN_ENABLED()
-- Triggered when player exits combat / loses aggro.
-- Let's generate stats screen, if enabled, and only if we took damage. (Spam otherwise)
--------------------------------------------------------------------------------
function ArdentDefender:PLAYER_REGEN_ENABLED()

	if ( AD_Info.lastCombat.DamageTaken ~= 0 and addon.db.profile.combatAnnounce ) then
		addon:Print(addon:FormatStats("COMBAT",nil))
 	end
	
	
end

--------------------------------------------------------------------------------
-- addon:PLAYER_ENTERING_WORLD()
-- Triggered when the player loads the game.
-- This just checks to make sure you are a Paladin, and disables itself if you aren't,
-- also refreshes talent check.. but do we need to?
--------------------------------------------------------------------------------
function ArdentDefender:PLAYER_ENTERING_WORLD()
	local _, class = UnitClass("player")
	
		if (class ~= "PALADIN") then
			return
		end
		-- Version check. Resets the database back to defaults if youre version is different than the newest release. (Fixes some bugs)
	if ( not addon.db.profile.version or (addon.db.profile.version < defaults.profile.version)) then
		addon.db:ResetDB()
		addon:Print("Database outdated. Resetting defaults.")
		addon:Debug("Profile reset")
	end
	
	if (addon.db.profile.annoyRF and not HasRF()) then
		addon:RFAnnoy()
	end
	
end

function ArdentDefender:Output(t,c,target)
	local isParty = ((GetNumSubgroupMembers() >0) and not IsInRaid())
	local isOfficer = true
	local isLFR = IsInRaid(LE_PARTY_CATEGORY_INSTANCE)
	
	if (isLFR) then
		if (UnitIsGroupAssistant("player",LE_PARTY_CATEGORY_INSTANCE) == true or UnitIsGroupLeader("player",LE_PARTY_CATEGORY_INSTANCE) == true) then
			isOfficer = true
		else
			isOfficer = false
		end
	elseif (IsInRaid(LE_PARTY_CATEGORY_HOME) and not isLFR) then
		if (UnitIsGroupAssistant("player",LE_PARTY_CATEGORY_HOME) == true or UnitIsGroupLeader("player",LE_PARTY_CATEGORY_HOME) == true) then
			isOfficer = true
		else
			isOfficer = false
		end
	end


	if not t then
		addon:Debug("Output: Null message")
		return
	end
	if not c then
		addon:Debug("Output: Null channel")
		return
	end
	addon:Debug("Output: Called with: "..t)
	local s = format(">>>> %s <<<<",string.gsub(t,"<target>",target))
	
	if (addon.db.profile.quietMode == true) then -- If quiet mode is emabled, any enabled output is redirectd to local instead. Only works if the ability is set to announce somewhere.
		local enabled = false
		for k, v in pairs(c) do
			if (k and v == true) then
				enabled = true
			end
		end
		if (enabled == true) then
			addon:Print(string.gsub(s,"<you>",target))
		end
		return
	end
	
	if (c["WHISPER"] == true  and target) then
		SendChatMessage(s,"WHISPER",nil,target)
	end
	
	s = string.gsub(s,"<you>",target) -- This is specifically for Hand of xxx messages, so that they say <you> when whispered, and <targetname> when announced.
	
	
	if (c["SAY"] == true) then
		SendChatMessage(s,"SAY")
	end
	if (c["YELL"] == true) then
		SendChatMessage(s,"YELL")
	end
	if (c["LOCAL"] == true) then
		addon:Print(t)
	end
	if (c["PARTY"] == true and isParty) then
		if (IsInGroup(LE_PARTY_CATEGORY_HOME)) then
			SendChatMessage(s,"PARTY")
		else
			SendChatMessage(s,"INSTANCE_CHAT")
		end
	end

	if (c["RAID"] == true) then
		if (IsInRaid(LE_PARTY_CATEGORY_HOME)) then
			SendChatMessage(s,"RAID")
		elseif (IsInRaid(LE_PARTY_CATEGORY_INSTANCE)) then
			SendChatMessage(s,"INSTANCE_CHAT")
		end
	end
	
	if (c["RAID_WARNING"] == true and isOfficer == true and not IsInRaid(LE_PARTY_CATEGORY_INSTANCE)) then
		SendChatMessage(s,"RAID_WARNING")
	end
	
end

--------------------------------------------------------------------------------
-- Formats either Session or last Combat info into a standard string for showing
-- returns a string
--------------------------------------------------------------------------------
function ArdentDefender:FormatStats(type,channel)
	local str, i = nil
	local totalDmg = 0
	local prewrapper, endwrapper, Averted, Absorbed, Blocked, Healed
	local Dodged,Missed,Parried

	if (type == "SESSION") then
		i = AD_Info.session
	else
		i = AD_Info.lastCombat
	end
	
	if (channel == nil) then  -- This is such an ugly kludge, but we can't sent color or multiline messages to chat channels.
		prewrapper = BLU.."\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|r\n"
		endwrapper = BLU.."\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|r"
		Averted, Absorbed, Blocked, Healed = RED.."Zero|r",RED.."Zero|r",RED.."Zero|r",RED.."Zero|r"
		Dodged,Missed,Parried = RED.."Zero|r", RED.."Zero|r", RED.."Zero|r"
	else 
		Averted, Absorbed, Blocked, Healed = "Zero","Zero","Zero","Zero"
		Dodged,Missed,Parried = "Zero","Zero","Zero"
	end
	
	totalDmg = i.DmgAbsorbed + i.DmgBlocked
	if (i.DeathsAverted > 0) then
		Averted = i.DeathsAverted
	end
	if (i.DmgAbsorbed >0) then
		Absorbed = i.DmgAbsorbed
	end
	if (i.DmgBlocked > 0) then
		Blocked = i.DmgBlocked
	end
	if (i.AmtHealed > 0) then
		Healed = i.AmtHealed
	end
	if (i.hitsDodged > 0) then
		Dodged = i.hitsDodged
	end
	
	if (i.hitsParried > 0) then
		Parried = i.hitsParried
	end
	if (i.hitsMissed > 0) then
		Missed = i.hitsMissed
	end
	if (channel == nil) then
		str = format ("%s%sDeaths Averted: |r%s\n%sBlocked:|r %s|r    %sAbsorbed: |r%s\n",prewrapper,GRN,Averted,GRN,Blocked,GRN,Absorbed )
		str = str..format("%sHits Dodged:|r%s   %sHits Parried: |r%s  %sHits Missed: |r%s\n", GRN,Dodged,GRN,Parried,GRN,Missed)
		str = str..format("%sArdent Heals:|r %s\n%sDmg Mitigated:|r %d\n%sDmg Taken: |r%d%s",GRN,Healed,GRN,totalDmg,GRN,i.DamageTaken,endwrapper)
		return str
	else
		SendChatMessage("Ardent Defender statistics",channel)
		SendChatMessage(format("Deaths Averted: %s, Blocked: %s, Absorbed: %s",Averted,Blocked,Absorbed),channel)
		SendChatMessage(format("Ardent Heals: %s, Dmg Mitigated: %d, Dmg Taken: %d", Healed,totalDmg,i.DamageTaken),channel)
		return
	end
end

--
--------------------------------------------------------------------------------
-- Basic announce function for all events we need to announce
--------------------------------------------------------------------------------
function ArdentDefender:Announce(type,event,special)

	local msg = nil
	local status = (event) and "UP!!!" or "DOWN!!!"
	local Flags = nil
	local output = nil
	local pvp = false
	addon:Debug("Announce() called with arguments: ", type or "nil", "/",event, "-",special or "nil")
	
	
	local pvpType, isFFA, faction = GetZonePVPInfo()
	if pvpType == "combat"  then  -- Tol Barad or Wintergrasp
		pvp = true
	end
	
	-- Any other instanced BG
	SetMapToCurrentZone()
	_, instance = IsInInstance()
	if (instance == "pvp" ) then
		pvp = true
	end
	
	if (type == "AVERT" and event) then
		
		msg = string.gsub(addon.db.profile.avert.messageUp,"<heal>",tostring(special))
		output = addon.db.profile.avert.output
		if ( addon.db.profile.specialFX) then
			PlaySoundFile(AD_SOUND)
			addon:Flash()
			addon:Shake()
		end
	elseif (type == "AVENGERMISS" and not event and addon.db.profile.avengerTracker ~= "DISABLED") then
		msg = addon.db.profile.avengerMessage
		if (addon.db.profile.avengerTracker == "LOCAL") then
			addon:Output(msg, {["LOCAL"] = true}, special)
		elseif (addon.db.profile.avengerTracker == "RAID") then
			addon:Output(msg, {["RAID"] = true}, special)
		elseif (addon.db.profile.avengerTracker == "PARTY") then
			addon:Output(msg, {["PARTY"] = true}, special)
		elseif (addon.db.profile.avengerTracker == "YELL") then
			addon:Output(msg, {["YELL"] = true}, special)
		end
		return
		
	elseif (type == "BUBBLEWALL" ) then
		if (event) then
			msg = addon.db.profile.bubbleWall.messageUp
		else
			msg = addon.db.profile.bubbleWall.messageDown
		end
		output = addon.db.profile.bubbleWall.output
	elseif (type == "TAUNTMISS") then
		if (event) then
			msg = "Target is IMMUNE to taunt!"
		else
			msg = "Taunt resisted!"
		end
		output = addon.db.profile.tauntOutput
	elseif (type == "OTHERPARRY" and event == true) then -- Boss parried someone else's attack.. why are they standing in front??
		if (addon.db.profile.parryTracker == "WHISPER" and event == true) then  -- If enabled, lets whisper them
			msg = "(AutoMessage) You caused my target to PARRY, stand behind it!"
			SendChatMessage(msg, "WHISPER", nil, special)
			return
		elseif (addon.db.profile.parryTracker == "LOCAL") then      -- Otherwise we print to local
			msg = special.." has caused your target to PARRY!!"
			addon:Output(msg,{["LOCAL"] = true},special)
			return
		end
	elseif (type == "DEVAURA" ) then -- Devotion Aura
				
		if (event) then
			msg = addon.db.profile.devAura.messageUp
		else
		
			msg = addon.db.profile.devAura.messageDown
		end
		
		output = addon.db.profile.devAura.output
	elseif (type == "HANDSAC" ) then
	    if (event) then
			msg = addon.db.profile.handSac.messageUp
		else
			msg = addon.db.profile.handSac.messageDown
		end
		output = addon.db.profile.handSac.output
	elseif (type == "SALV" and event ) then
		msg = addon.db.profile.handSalv.messageUp
		output = addon.db.profile.handSalv.output
		if (addon.db.profile.basicMode == true) then
			addon:Output(msg,{["WHISPER"] = true},special)
			return
		end
	elseif (type == "HOP" and event ) then
		msg = addon.db.profile.handProt.messageUp
		output = addon.db.profile.handProt.output
		if (addon.db.profile.basicMode == true) then
			addon:Output(msg,{["WHISPER"] = true},special)
			return
		end
	elseif (type == "DI" and event) then
		msg = addon.db.profile.di.messageUp
		output = addon.db.profile.di.output
		if (addon.db.profile.basicMode == true) then
			addon:Output(msg,{["WHISPER"] = true},special)
			return
		end
	elseif (type == "LOH" and event) then
		msg = addon.db.profile.loh.messageUp
		output = addon.db.profile.loh.output
		
		if (addon.db.profile.basicMode == true) then
			addon:Output(msg,{["WHISPER"] = true},special)
			return
		end
	elseif (type == "FREEDOM" and event) then
		msg = addon.db.profile.handfree.messageUp
		
		if (addon.db.profile.basicMode == true) then
			addon:Output(msg,{["WHISPER"] = true},special)
			return
		end
		output = addon.db.profile.handfree.output
	elseif (type == "RFHACK" and addon.db.profile.annoyRF and not HasRF()) then
		addon:RFAnnoy()
	elseif (type == "OTHERTAUNT" ) then
		if (addon.db.profile.otherTaunt== false) then
			addon:Output(special.." has TAUNTED your target with "..event,{["LOCAL"] = true}, special)
			return
		else
			output = addon.db.profile.tauntOutput
			msg = special.." has TAUNTED my target with "..event
		end
		
	elseif (type == "ANCIENTGUARDIAN") then
		if (event) then
			msg = addon.db.profile.ancientGuardian.messageUp
		else
			msg = addon.db.profile.ancientGuardian.messageDown
		end
		output = addon.db.profile.ancientGuardian.output
	elseif (type == "ARDENTDEFENDER") then
		if (event ) then
			msg =addon.db.profile.ardentDefender.messageUp
		else
			msg = addon.db.profile.ardentDefender.messageDown
		end
		output = addon.db.profile.ardentDefender.output
	else
	    --addon:Debug("Announce() invalid argument! "..type)
	    msg = nil
	end
	if (not output) then
		self:Debug("Announce: Nil output")
		return
	end
	if (addon.db.profile.basicMode == true) then
		addon:Debug("Announce: basicmode is on.")
		output = addon.db.profile.tauntOutput
	end
	if (addon.db.profile.onlyTanking == true and not IsProt()) then
		return
	end
	if (msg and not pvp) then
		addon:Output(msg,output,special)
	elseif (msg and pvp) then
		addon:Output(msg,{["LOCAL"] = true},special)
	end
end

-- Creates a frame, attaches our texture and flashes it, this is adapted from the Omen addon
function ArdentDefender:Flash()
	if not addon.FlashFrame then
		local flasher = CreateFrame("Frame", "ADFlashFrame")
		flasher:SetToplevel(true)
		flasher:SetFrameStrata("FULLSCREEN_DIALOG")
		flasher:SetAllPoints(UIParent)
		flasher:EnableMouse(false)
		flasher:Hide()
		flasher.texture = flasher:CreateTexture(nil, "BACKGROUND")
		flasher.texture:SetTexture(AD_FLASHTEXTURE)
		flasher.texture:SetAllPoints(UIParent)
		flasher.texture:SetBlendMode("ADD")
		flasher:SetScript("OnShow", function(self)
			self.elapsed = 0
			self:SetAlpha(0)
		end)
		flasher:SetScript("OnUpdate", function(self, elapsed)
			elapsed = self.elapsed + elapsed
			if elapsed < 2.6 then
				local alpha = elapsed % 1.3
				if alpha < 0.15 then
					self:SetAlpha(alpha / 0.15)
				elseif alpha < 0.9 then
					self:SetAlpha(1 - (alpha - 0.15) / 0.6)
				else
					self:SetAlpha(0)
				end
			else
				self:Hide()
			end
			self.elapsed = elapsed
		end)
		addon.FlashFrame = flasher
	end

	addon.FlashFrame:Show()
end

-- This was adapted from Omen3 (with permission), their original comments are below. --Blind
-- This function is adapted from Omen2 to be self-contained,
-- which was initially taken from BigWigs
function ArdentDefender:Shake()
	local shaker = addon.ShakerFrame
	if not shaker then
		shaker = CreateFrame("Frame", "ADShaker", UIParent)
		shaker:Hide()
		shaker:SetScript("OnUpdate", function(self, elapsed)
			elapsed = self.elapsed + elapsed
			local x, y = 0, 0 -- Resets to original position if we're supposed to stop.
			if elapsed >= 0.8 then
				self:Hide()
			else
				x, y = random(-8, 8), random(-8, 8)
			end
			if (WorldFrame:IsProtected() and InCombatLockdown()) then
				if not shaker.fail then
					addon:Output("Cannot shake world frame if you have had nameplates turned on at least once since logging in.","LOCAL",{["LOCAL"] = true},"nil")
					shaker.fail = true
				end
				self:Hide()
			else
				WorldFrame:ClearAllPoints()
				for i = 1, #self.originalPoints do
					local v = self.originalPoints[i]
					WorldFrame:SetPoint(v[1], v[2], v[3], v[4] + x, v[5] + y)
				end
			end
			self.elapsed = elapsed
		end)
		shaker:SetScript("OnShow", function(self)
			-- Store old worldframe positions, we need them all, people have frame modifiers for it
			if not self.originalPoints then
				self.originalPoints = {}
				for i = 1, WorldFrame:GetNumPoints() do
					tinsert(self.originalPoints, {WorldFrame:GetPoint(i)})
				end
			end
			self.elapsed = 0
		end)
		addon.ShakerFrame = shaker
	end

	shaker:Show()
end
function HasRF()
	local n = UnitAura("player","Righteous Fury")
	
	
	if (IsProt() and not n) then
		return false
	else
		return true
	end
end
function ArdentDefender:RFAnnoy()
	if not addon.RFFrame then
		local frame = CreateFrame("Frame", "ADAnnoyFrame")
		frame:SetToplevel(true)
		frame:SetFrameStrata("FULLSCREEN_DIALOG")
		frame:SetAllPoints(UIParent)
		frame:EnableMouse(false)
		frame:Hide()
		frame.texture = frame:CreateTexture(nil, "BACKGROUND")
		frame.texture:SetTexture("Interface\\FullScreenTextures\\LowHealth")
		frame.texture:SetAllPoints(UIParent)
		frame.texture:SetBlendMode("ADD")
		frame:SetScript("OnShow", function(self)
			self.elapsed = 0
			self:SetAlpha(0)
		end)
		frame:SetScript("OnUpdate", function(self, elapsed)
			elapsed = self.elapsed + elapsed
			if (HasRF() or not ArdentDefender.db.profile.annoyRF) then
				self:Hide()
			end
			if elapsed < 2.6 then
				local alpha = elapsed % 1.3
				if alpha < 0.15 then
					self:SetAlpha(alpha / 0.15)
				elseif alpha < 0.9 then
					self:SetAlpha(1 - (alpha - 0.15) / 0.6)
				else
					self:SetAlpha(0)
				end
			else
				elapsed = 0
				
			end
			self.elapsed = elapsed
		end)
		addon.RFFrame = frame
	end

	addon.RFFrame:Show()
end

local function GetNextTag()
	local i = 8
	
	while i > 0 do
		if (RaidTags[i] == nil) then
			return i
		end
	i = i - 1
	end
  
      return nil
end

local function UnitIsEligable (unit)
	local type = UnitCreatureType(unit)
	return UnitExists(unit)	and (UnitCanAttack("player", unit) or UnitIsEnemy("player", unit)) and not UnitIsDead(unit) and  type ~= "Critter" and type ~= "Totem" and not UnitPlayerControlled(unit)  and not UnitIsPlayer(unit)
end

function ArdentDefender:ResetTags()
	for i = 1,8 do
		RaidTags[i] = nil
		SetRaidTargetIcon("player",i)
	end
	SetRaidTargetIcon("player",0)
end

		
function ArdentDefender:UPDATE_MOUSEOVER_UNIT()
	local unit = "mouseover"
	local tag = GetNextTag()
	
	if (addon.db.profile.quickMarks ~= true) then return
	end
	
	if ( IsAltKeyDown() and IsControlKeyDown()) then 
		if (UnitIsEligable(unit) ~= true or  GetRaidTargetIndex(unit) ~=nil or tag == nil) then return end
		SetRaidTarget(unit,tag)
		RaidTags[tag] =  UnitGUID(unit)

	end
end
